@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
@@ -14,12 +12,16 @@
 
 from collections import OrderedDict
 import os
+import random
+import re
 
 from docutils import nodes
+from docutils.parsers import rst
 from docutils.parsers.rst.directives.tables import Table
 from docutils.statemachine import ViewList
 import pbr.version
-from sphinx.util.compat import Directive
+import six
+from sphinx.util import logging
 from sphinx.util.osutil import copyfile
 import yaml
 
@@ -30,6 +32,8 @@ from os_api_ref.http_codes import HTTPResponseCodeDirective
 __version__ = pbr.version.VersionInfo(
     'os_api_ref').version_string()
 
+LOG = logging.getLogger(__name__)
+
 # This is to allow for a graceful swap from oslosphinx to openstackdocstheme
 THEME = 'openstackdocstheme'
 
@@ -66,7 +70,8 @@ in both naming and ordering of parameters at every declaration.
 """
 
 
-def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
+def ordered_load(
+        stream, Loader=yaml.SafeLoader, object_pairs_hook=OrderedDict):
     """Load yaml as an ordered dict
 
     This allows us to inspect the order of the file on disk to make
@@ -123,7 +128,7 @@ class rest_expand_all(nodes.Part, nodes.Element):
     pass
 
 
-class RestExpandAllDirective(Directive):
+class RestExpandAllDirective(rst.Directive):
     # This tells sphinx that the directive will need to generate
     # content during the final build phase.
     has_content = True
@@ -133,12 +138,14 @@ class RestExpandAllDirective(Directive):
         node = rest_expand_all()
         max_ver = app.config.os_api_ref_max_microversion
         min_ver = app.config.os_api_ref_min_microversion
+        releases = app.config.os_api_ref_release_microversions
         node['major'] = None
         try:
             if max_ver.split('.')[0] == min_ver.split('.')[0]:
                 node['max_ver'] = int(max_ver.split('.')[1])
                 node['min_ver'] = int(min_ver.split('.')[1])
                 node['major'] = int(max_ver.split('.')[0])
+                node['releases'] = releases
         except ValueError:
             # TODO(sdague): warn that we're ignoring this all
             pass
@@ -147,7 +154,7 @@ class RestExpandAllDirective(Directive):
         return [node]
 
 
-class RestMethodDirective(Directive):
+class RestMethodDirective(rst.Directive):
 
     # this enables content in the directive
     has_content = True
@@ -164,8 +171,6 @@ class RestMethodDirective(Directive):
         lineno = self.state_machine.abs_line_number()
         target = nodes.target()
         section = nodes.section(classes=["detail-control"])
-        # env = self.state.document.settings.env
-        # env.app.info("Parent %s" % self.state.parent.attributes)
 
         node = rest_method()
 
@@ -177,6 +182,12 @@ class RestMethodDirective(Directive):
 
         node['method'] = method
         node['url'] = url
+
+        # Extract the path parameters from the url
+        env = self.state.document.settings.env
+        env.path_params = []
+        env.path_params = re.findall("{[a-zA-Z][a-zA-Z_0-9]*}", url)
+
         node['target'] = self.state.parent.attributes['ids'][0]
         node['css_classes'] = ""
         if node['min_version']:
@@ -188,7 +199,8 @@ class RestMethodDirective(Directive):
 
         # We need to build a temporary target that we can replace
         # later in the processing to get the TOC to resolve correctly.
-        temp_target = "%s-selector" % node['target']
+        temp_target = "%s-%d-selector" % (node['target'],
+                                          random.randint(1, 1000))
         target = nodes.target(ids=[temp_target])
         self.state.add_target(temp_target, '', target, lineno)
         section += node
@@ -210,25 +222,22 @@ class RestParametersDirective(Table):
             return YAML_CACHE[fpath]
 
         lookup = {}
-        # self.app.info("Fpath: %s" % fpath)
         try:
             with open(fpath, 'r') as stream:
                 lookup = ordered_load(stream)
         except IOError:
-            self.env.warn(
-                self.env.docname,
-                "Parameters file %s not found" % fpath)
+            LOG.warning("Parameters file not found, %s", fpath,
+                        location=(self.env.docname, None))
             return
         except yaml.YAMLError as exc:
-            self.app.warn(exc)
+            LOG.exception(exc_info=exc)
             raise
 
         if lookup:
             self._check_yaml_sorting(fpath, lookup)
         else:
-            self.env.warn(
-                self.env.docname,
-                "Parameters file is empty %s" % fpath)
+            LOG.warning("Parameters file is empty, %s", fpath,
+                        location=(self.env.docname, None))
             return
 
         YAML_CACHE[fpath] = lookup
@@ -249,15 +258,18 @@ class RestParametersDirective(Table):
 
         last = None
         for key, value in yaml_data.items():
+            if not isinstance(value, dict):
+                raise Exception('Expected a dict for {0}; got {0}={1}).\n'
+                                'You probably have indentation typo in your'
+                                'YAML source'.format(key, value))
+
             # use of an invalid 'in' value
             if value['in'] not in sections:
-                self.app.warn(
-                    "``%s`` is not a valid value for 'in' (must be one of: %s)"
-                    ". (see ``%s``)" % (
-                        value['in'],
-                        ", ".join(sorted(sections.keys())),
-                        key)
-                    )
+                LOG.warning("``%s`` is not a valid value for 'in' (must be "
+                            "one of: %s). (see ``%s``)",
+                            value['in'],
+                            ", ".join(sorted(sections.keys())),
+                            key)
                 continue
 
             if last is None:
@@ -267,17 +279,13 @@ class RestParametersDirective(Table):
             current_section = value['in']
             last_section = last[1]['in']
             if sections[current_section] < sections[last_section]:
-                self.app.warn(
-                    "Section out of order. All parameters in section ``%s`` "
-                    "should be after section ``%s``. (see ``%s``)" % (
-                        last_section,
-                        current_section,
-                        last[0]))
+                LOG.warning("Section out of order. All parameters in section "
+                            "``%s`` should be after section ``%s``. (see "
+                            "``%s``)", last_section, current_section, last[0])
             if (sections[value['in']] == sections[last[1]['in']] and
                 key.lower() < last[0].lower()):
-                self.app.warn(
-                    "Parameters out of order ``%s`` should be after ``%s``" % (
-                        last[0], key))
+                LOG.warning("Parameters out of order ``%s`` should be after "
+                            "``%s``", last[0], key)
             last = (key, value)
 
     def yaml_from_file(self, fpath):
@@ -292,42 +300,51 @@ class RestParametersDirective(Table):
             return
 
         content = "\n".join(self.content)
-        parsed = yaml.load(content)
-        # self.app.info("Params loaded is %s" % parsed)
-        # self.app.info("Lookup table looks like %s" % lookup)
+        parsed = yaml.safe_load(content)
         new_content = list()
         for paramlist in parsed:
             if not isinstance(paramlist, dict):
-                self.env.warn(
-                    "%s:%s" % (
-                        self.state_machine.node.source,
-                        self.state_machine.node.line),
-                    ("Invalid parameter definition ``%s``. Expected "
-                     "format: ``name: reference``. "
-                     " Skipping." % paramlist))
+                location = (self.state_machine.node.source,
+                            self.state_machine.node.line)
+                LOG.warning("Invalid parameter definition ``%s``. Expected "
+                            "format: ``name: reference``. Skipping.",
+                            paramlist, location=location)
                 continue
             for name, ref in paramlist.items():
                 if ref in lookup:
                     new_content.append((name, lookup[ref]))
                 else:
                     # TODO(sdague): this provides a kind of confusing
-                    # error message because env.warn isn't meant to be
+                    # error message because app.warn isn't meant to be
                     # used this way, however it does provide a way to
                     # track down where the parameters list is that is
                     # wrong. So it's good enough for now.
-                    self.env.warn(
-                        "%s:%s " % (
-                            self.state_machine.node.source,
-                            self.state_machine.node.line),
-                        ("No field definition for ``%s`` found in ``%s``. "
-                         " Skipping." % (ref, fpath)))
+                    location = (self.state_machine.node.source,
+                                self.state_machine.node.line)
+                    LOG.warning("No field definition for ``%s`` found in "
+                                "``%s``. Skipping.", ref, fpath)
+
+                # Check for path params in stanza
+                for i, param in enumerate(self.env.path_params):
+                    if (param.rstrip('}').lstrip('{')) == name:
+                        del self.env.path_params[i]
+                        break
+                    else:
+                        continue
+
+        if len(self.env.path_params) is not 0:
+            # Warn that path parameters are not set in rest_parameter
+            # stanza and will not appear in the generated table.
+            for param in self.env.path_params:
+                location = (self.state_machine.node.source,
+                            self.state_machine.node.line)
+                LOG.warning("No path parameter ``%s`` found in rest_parameter"
+                            " stanza.\n", param.rstrip('}').lstrip('{'))
 
-        # self.app.info("New content %s" % new_content)
         self.yaml = new_content
 
     def run(self):
         self.env = self.state.document.settings.env
-        self.app = self.env.app
 
         # Make sure we have some content, which should be yaml that
         # defines some parameters.
@@ -355,11 +372,15 @@ class RestParametersDirective(Table):
         # TODO(sdague): it would be good to dynamically set column
         # widths (or basically make the colwidth thing go away
         # entirely)
-        self.options['widths'] = (20, 10, 10, 60)
+        self.options['widths'] = [20, 10, 10, 60]
         self.col_widths = self.get_column_widths(self.max_cols)
+        if isinstance(self.col_widths, tuple):
+            # In docutils 0.13.1, get_column_widths returns a (widths,
+            # colwidths) tuple, where widths is a string (i.e. 'auto').
+            # See https://sourceforge.net/p/docutils/patches/120/.
+            self.col_widths = self.col_widths[1]
         # Actually convert the yaml
         title, messages = self.make_title()
-        # self.app.info("Title %s, messages %s" % (title, messages))
         table_node = self.build_table()
         self.add_name(table_node)
         if title:
@@ -371,7 +392,7 @@ class RestParametersDirective(Table):
         groups = []
         trow = nodes.row()
         entry = nodes.entry()
-        para = nodes.paragraph(text=unicode(table_data))
+        para = nodes.paragraph(text=six.u(table_data))
         entry += para
         trow += entry
         rows.append(trow)
@@ -399,7 +420,6 @@ class RestParametersDirective(Table):
         rows = []
         groups = []
         try:
-            # self.app.info("Parsed content is: %s" % self.yaml)
             for key, values in self.yaml:
                 min_version = values.get('min_version', '')
                 max_version = values.get('max_version', '')
@@ -411,7 +431,7 @@ class RestParametersDirective(Table):
                                         str(min_version).replace('.', '_'))
                     classes.append(min_ver_css_name)
                 if max_version:
-                    desc += ("\n\n**Deprecated in version %s**\n" %
+                    desc += ("\n\n**Available until version %s**\n" %
                              max_version)
                     max_ver_css_name = ("rp_max_ver_" +
                                         str(max_version).replace('.', '_'))
@@ -427,8 +447,8 @@ class RestParametersDirective(Table):
                 rows.append(trow)
         except AttributeError as exc:
             if 'key' in locals():
-                self.app.warn("Failure on key: %s, values: %s. %s" %
-                              (key, values, exc))
+                LOG.warning("Failure on key: %s, values: %s. %s",
+                            key, values, exc)
             else:
                 rows.append(self.show_no_yaml_error())
         return rows, groups
@@ -471,7 +491,9 @@ def rest_method_html(self, node):
 <div class="row">
     <div class="col-md-2">
     <div class="operation">
-    <a name="%(target)s" class="operation-anchor" href="#%(target)s">
+    <a name="%(target)s" class="operation-anchor" href="#%(target)s"
+      onclick="window.location.hash = hash;"
+      >
       <span class="glyphicon glyphicon-link"></span></a>
     <span class="label label-%(method)s">%(method)s</span>
     </div>
@@ -510,7 +532,9 @@ def rest_expand_all_html(self, node):
     tmpl = """
 <div class="row">
 %(extra_js)s
-<div class=col-md-11>%(selector)s</div>
+<div class="col-md-2 col-md-offset-9">
+%(selector)s
+</div>
 <div class=col-md-1>
     <button id="expand-all"
        data-toggle="collapse"
@@ -519,28 +543,64 @@ def rest_expand_all_html(self, node):
 </div>
 </div>"""
 
+    node.setdefault('selector', "")
+    node.setdefault('extra_js', "")
+
     if node['major']:
-        selector = """
-<div class="btn-group" role="group" aria-label="...">
-    <button type="button"
-        class="btn btn-default mv_selector active">All</button>\n"""
-        for x in range(node['min_ver'], node['max_ver'] + 1):
-            selector += ('<button type="button" class="btn btn-default '
-                         'mv_selector">%d.%d</button>\n' % (node['major'], x))
-        selector += "</div>"
-        node['selector'] = selector
-        node['extra_js'] = ("<script>os_max_mv = %d; "
-                            "os_min_mv = %d;</script>") % (
-                                node['max_ver'],
-                                node['min_ver'])
-    else:
-        node['selector'] = ""
-        node['extra_js'] = ""
+        node['selector'], node['extra_js'] = create_mv_selector(node)
 
     self.body.append(tmpl % node)
     raise nodes.SkipNode
 
 
+def create_mv_selector(node):
+
+    mv_list = '<option value="" selected="selected">All</option>'
+
+    for x in range(node['min_ver'], node['max_ver'] + 1):
+        mv_list += build_mv_item(node['major'], x, node['releases'])
+
+    selector_tmpl = """
+<form class=form-inline">
+<div class="form-group">
+<label class="control-label">
+Microversions
+</label>
+<select class="combobox form-control" id="mv_select">
+    %(mv_list)s
+</select>
+</div>
+</form>
+"""
+
+    js_tmpl = """
+<script>
+    os_max_mv = %(max)d;
+    os_min_mv = %(min)d;
+</script>
+"""
+
+    selector_content = {
+        'mv_list': mv_list
+    }
+
+    js_content = {
+        'min': node['min_ver'],
+        'max': node['max_ver']
+    }
+
+    return selector_tmpl % selector_content, js_tmpl % js_content
+
+
+def build_mv_item(major, micro, releases):
+    version = "%d.%d" % (major, micro)
+    if version in releases:
+        return '<option value="%s">%s - %s</option>' % (
+            version, version, releases[version].capitalize())
+    else:
+        return '<option value="%s">%s</option>' % (version, version)
+
+
 def resolve_rest_references(app, doctree):
     for node in doctree.traverse():
         if isinstance(node, rest_method):
@@ -574,16 +634,16 @@ def resolve_rest_references(app, doctree):
 
 
 def copy_assets(app, exception):
-    assets = ('api-site.css', 'api-site.js')
+    assets = ('api-site.css', 'api-site.js', 'combobox.js')
     fonts = (
         'glyphicons-halflings-regular.ttf',
         'glyphicons-halflings-regular.woff'
     )
-
-    if app.builder.name != 'html' or exception:
+    builders = ('html', 'readthedocs', 'readthedocssinglehtmllocalmedia')
+    if app.builder.name not in builders or exception:
         return
-    app.info('Copying assets: %s' % ', '.join(assets))
-    app.info('Copying fonts: %s' % ', '.join(fonts))
+    LOG.info('Copying assets: %s', ', '.join(assets))
+    LOG.info('Copying fonts: %s', ', '.join(fonts))
     for asset in assets:
         dest = os.path.join(app.builder.outdir, '_static', asset)
         source = os.path.abspath(os.path.dirname(__file__))
@@ -597,21 +657,20 @@ def copy_assets(app, exception):
 def add_assets(app):
     app.add_stylesheet('api-site.css')
     app.add_javascript('api-site.js')
+    app.add_javascript('combobox.js')
 
 
 def setup(app):
     # Add some config options around microversions
     app.add_config_value('os_api_ref_max_microversion', '', 'env')
     app.add_config_value('os_api_ref_min_microversion', '', 'env')
+    app.add_config_value('os_api_ref_release_microversions', '', 'env')
     # TODO(sdague): if someone wants to support latex/pdf, or man page
     # generation using these stanzas, here is where you'd need to
     # specify content specific renderers.
-    app.add_node(rest_method,
-                 html=(rest_method_html, None))
-    app.add_node(rest_expand_all,
-                 html=(rest_expand_all_html, None))
-    app.add_node(http_code,
-                 html=(http_code_html, None))
+    app.add_node(rest_method, html=(rest_method_html, None))
+    app.add_node(rest_expand_all, html=(rest_expand_all_html, None))
+    app.add_node(http_code, html=(http_code_html, None))
 
     # This specifies all our directives that we're adding
     app.add_directive('rest_parameters', RestParametersDirective)
