@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
@@ -20,11 +18,26 @@ Tests for `os_api_ref` module.
 """
 
 from bs4 import BeautifulSoup
+import sphinx
 from sphinx_testing import with_app
 
 from os_api_ref.tests import base
 
 
+# FIXME(stephenfin): This is horrible. We're monkeypatching this to work around
+# the fact that Sphinx 1.8+ started called 'abspath' from within the
+# 'sphinx.application.Application' class [1]. This means our careful use of
+# 'sphinx_testing.path.path' for 'Application.outdir' etc. gets stomped on.
+# We're correcting that but we're doing so globally because mock doesn't work
+# for some reason and this is bound to have some side effects
+#
+# [1] https://github.com/sphinx-doc/sphinx/commit/3a85b3502f
+try:
+    sphinx.application.abspath = lambda x: x
+except ImportError:  # Sphinx < 1.8
+    pass
+
+
 class TestBasicExample(base.TestCase):
     """Test basic rendering.
 
@@ -40,7 +53,7 @@ class TestBasicExample(base.TestCase):
         self.status = status
         self.warning = warning
         self.app.build()
-        self.html = (app.outdir / 'index.html').read_text()
+        self.html = (app.outdir / 'index.html').read_text(encoding='utf-8')
         self.soup = BeautifulSoup(self.html, 'html.parser')
         self.content = str(self.soup)
 
@@ -80,10 +93,10 @@ class TestBasicExample(base.TestCase):
 
         table = """<table border="1" class="docutils">
 <colgroup>
-<col width="20%"></col>
-<col width="10%"></col>
-<col width="10%"></col>
-<col width="60%"></col>
+<col width="20%"/>
+<col width="10%"/>
+<col width="10%"/>
+<col width="60%"/>
 </colgroup>
 <thead valign="bottom">
 <tr class="row-odd"><th class="head">Name</th>
@@ -107,8 +120,8 @@ class TestBasicExample(base.TestCase):
 
         success_table = """table border="1" class="docutils">
 <colgroup>
-<col width="30%"></col>
-<col width="70%"></col>
+<col width="30%"/>
+<col width="70%"/>
 </colgroup>
 <thead valign="bottom">
 <tr class="row-odd"><th class="head">Code</th>
@@ -131,8 +144,8 @@ class TestBasicExample(base.TestCase):
 
         error_table = """<table border="1" class="docutils">
 <colgroup>
-<col width="30%"></col>
-<col width="70%"></col>
+<col width="30%"/>
+<col width="70%"/>
 </colgroup>
 <thead valign="bottom">
 <tr class="row-odd"><th class="head">Code</th>
